with Registers;
with System.Machine_Code;

package body Utils is

   procedure Enable_Interrupts is
   begin
      System.Machine_Code.Asm (Template => "cpsie i", Volatile => True);
   end Enable_Interrupts;
   
   procedure Disable_Interrupts is
   begin
      System.Machine_Code.Asm (Template => "cpsid i", Volatile => True);
   end Disable_Interrupts;

   procedure Wait_For_Interrupt is
   begin
      System.Machine_Code.Asm (Template => "wfi", Volatile => True);
   end Wait_For_Interrupt;

   Master_Clock_Frequency_Hz : constant := 8_000_000;
   SysTick_Divider : constant := 8;
   SysTick_Clock_Frequency_Hz : constant := Master_Clock_Frequency_Hz / SysTick_Divider;
   SysTick_Period_Millis : constant := 10;
   
   SysTick_Reload_Value : constant :=
      (SysTick_Clock_Frequency_Hz / 1_000) * SysTick_Period_Millis;

   System_Time : Time;
   pragma Volatile (System_Time);

   procedure SysTick_Handler is
   begin
      System_Time := System_Time + SysTick_Period_Millis;
   end SysTick_Handler;

   procedure Enable_System_Time is
      use type Registers.Word;
   begin
      System_Time := 0;
  
      Registers.SysTick_VAL := 0;
      Registers.SysTick_LOAD := SysTick_Reload_Value;
      Registers.SysTick_CTRL := Registers.Bit_0 or Registers.Bit_1;
   end Enable_System_Time;

   function Base_Tick return Time is
   begin
      return SysTick_Period_Millis;
   end Base_Tick;
   
   function Clock return Time is
   begin
      return System_Time;
   end Clock;

   procedure Reset_Clock (T : in Time := 0) is
   begin
      System_Time := T;
   end Reset_Clock;
   
   procedure Delay_Until (T : in Time) is
   begin
      while System_Time < T loop
         Wait_For_Interrupt;
      end loop;
   end Delay_Until;
   
   procedure Scheduler_Tick (Tasks : in out Task_Descriptions; Time_Passed : in Time) is
   begin
      for T in Tasks'Range loop
         if Tasks (T).Action /= null then
            Tasks (T).Time_Since_Last_Due_Active := Tasks (T).Time_Since_Last_Due_Active + Time_Passed;
            if Tasks (T).Time_Since_Last_Due_Active >= Tasks (T).Activation_Period then
               
               Tasks (T).Action.all;
               
               Tasks (T).Time_Since_Last_Due_Active := Tasks (T).Time_Since_Last_Due_Active - Tasks (T).Activation_Period;
            end if;
         end if;
      end loop;
   end Scheduler_Tick;
   
   -- workarounds for compiler-injected references which are useless here
   
   procedure Dummy_aeabi_unwind_cpp_pr0;
   pragma Export (C, Dummy_aeabi_unwind_cpp_pr0, "__aeabi_unwind_cpp_pr0");
   
   procedure Dummy_aeabi_unwind_cpp_pr1;
   pragma Export (C, Dummy_aeabi_unwind_cpp_pr1, "__aeabi_unwind_cpp_pr1");
   
   procedure Dummy_aeabi_unwind_cpp_pr0 is
   begin
      null;
   end Dummy_aeabi_unwind_cpp_pr0;
   
   procedure Dummy_aeabi_unwind_cpp_pr1 is
   begin
      null;
   end Dummy_aeabi_unwind_cpp_pr1;
   
   -- EABI functions
   
   function LMUL (A : in Registers.Long_Word; B : in Registers.Long_Word)
                 return Registers.Long_Word;
   pragma Export (C, LMUL, "__aeabi_lmul");
   
   function LMUL (A : in Registers.Long_Word; B : in Registers.Long_Word)
                 return Registers.Long_Word is
      use type Registers.Long_Word;
      
      Result : Registers.Long_Word := 0;
      Mask : Registers.Long_Word := 1 * 2**63;
   begin
      while Mask /= 0 loop
         Result := Result * 2;
         if (A and Mask) /= 0 then
            Result := Result + B;
         end if;
         Mask := Mask / 2;
      end loop;
      
      return Result;
   end LMUL;
   
end Utils;
